from ...managers import get_python_manager, get_remote_manager, get_source_table
from ...settings import get_mode
from ...FileManagment import TempFileManager
from ..parsers.RegionParser import RegionParser
import os
from . import FILES_FOLDER, SCHEMA_FILE, WEB_PROFILE_FILE, PROFILE_FILE


def get_schema_path(path):
    schema_path = os.path.join(path, SCHEMA_FILE)
    return schema_path


def get_file_paths(path):
    real_path = preprocess_path(path)
    all_files = os.listdir(real_path)

    def filter_files(x):
        return not (x in [SCHEMA_FILE, WEB_PROFILE_FILE, PROFILE_FILE] or x.startswith("_"))

    files_paths = list(map(lambda x: os.path.join(real_path, x), filter(filter_files, all_files)))
    # files_paths = set(glob.glob(real_path + "/[!_]*"))
    schema_path = get_schema_path(real_path)
    return files_paths, schema_path


def preprocess_path(path):
    """ Given a dataset path, the following structure is to be expected:
        - path/
            - files/
                - S_00000.gdm
                - S_00000.gdm.meta
                - S_00001.gdm
                - S_00001.gdm.meta
                - ...
                - schema.xml
                - [profile.xml]
                - [web_profile.xml]
            - [info.txt]
            - [query.txt]
            - [vocabulary.txt]

    :param path
    :return: the path where the gdm data are
    """
    for sub_f in os.listdir(path):
        sub_f_tot = os.path.join(path, sub_f)
        if os.path.isdir(sub_f_tot) and sub_f == FILES_FOLDER:
            if check_for_dataset(sub_f_tot):
                return sub_f_tot
            else:
                raise ValueError("Dataset in {} was not in GMQL format".format(sub_f_tot))
    # if we are here it means that there is no files folder...so we need to check the root
    if check_for_dataset(path):
        return path
    else:
        raise ValueError("Dataset in {} was not in GMQL format".format(path))


def check_for_dataset(files):
    """ A GDM dataset has the form:
        - S_00000.gdm
        - S_00000.gdm.meta
        - S_00001.gdm
        - S_00001.gdm.meta
        - ...
        - schema.xml
        - [profile.xml]
        - [web_profile.xml]
    :param files: path of the dataset
    :return: True if the path contains a gdm dataset
    """
    all_files = os.listdir(files)
    meta_files = set(map(lambda y: y[: -9], filter(lambda x: x.endswith(".gdm.meta"), all_files)))
    regs_files = set(map(lambda y: y[: -4], filter(lambda x: x.endswith(".gdm"), all_files)))
    return meta_files == regs_files


def load_from_file(path, parser: RegionParser):
    """ Loads a GDM dataset from a single BED-like file.

    :param path: location of the file
    :param parser: RegionParser object specifying the parser of the file
    :return: a GMQLDataset
    """
    from .. import GMQLDataset
    pmg = get_python_manager()
    id = add_to_sources(local_path=path, parser=parser)
    local_sources = [id]

    index = pmg.readFile(str(id), parser.get_gmql_parser())
    return GMQLDataset.GMQLDataset(index=index, parser=parser,
                                   location="local", path_or_name=path,
                                   local_sources=local_sources)


def load_from_path(local_path, parser=None):
    """ Loads the data from a local path into a GMQLDataset.
    The loading of the files is "lazy", which means that the files are loaded only when the
    user does a materialization (see :func:`~gmql.dataset.GMQLDataset.GMQLDataset.materialize` ).
    The user can force the materialization of the data (maybe for an initial data exploration on
    only the metadata) by setting the :attr:`~.reg_load` (load in memory the region data),
    :attr:`~.meta_load` (load in memory the metadata) or :attr:`~.all_load` (load both region and
    meta data in memory). If the user specifies this final parameter as True, a
    :class:`~gmql.dataset.GDataframe.GDataframe` is returned, otherwise a
    :class:`~gmql.dataset.GMQLDataset.GMQLDataset` is returned
    
    :param local_path: local path of the dataset
    :param parser: the parser to be used for reading the data
    :param all_load: if set to True, both region and meta data are loaded in memory and an 
                     instance of GDataframe is returned
    :return: A new GMQLDataset or a GDataframe
    """

    # from .. import GDataframe
    from .. import GMQLDataset
    pmg = get_python_manager()

    local_path = pmg.preProcessPath(local_path)

    # if all_load:
    #     # load directly the metadata for exploration
    #     meta = MetaLoaderFile.load_meta_from_path(local_path)
    #     if isinstance(parser, RegionParser):
    #         # region data
    #         regs = RegLoaderFile.load_reg_from_path(local_path, parser)
    #     else:
    #         regs = RegLoaderFile.load_reg_from_path(local_path)
    #
    #     return GDataframe.GDataframe(regs=regs, meta=meta)
    # else:
    # from ...settings import is_metaprofiling_enabled
    # if is_metaprofiling_enabled():
    #     meta_profile = create_metadata_profile(local_path)
    # else:
    meta_profile = None

    if parser is None:
        # find the parser
        parser = RegionParser.from_schema_file(local_path)
    elif not isinstance(parser, RegionParser):
        raise ValueError("parser must be RegionParser. {} was provided".format(type(parser)))

    id = add_to_sources(local_path=local_path, parser=parser)
    local_sources = [id]

    index = pmg.read_dataset(str(id), parser.get_gmql_parser())
    return GMQLDataset.GMQLDataset(index=index, parser=parser,
                                   location="local", path_or_name=local_path,
                                   local_sources=local_sources,
                                   meta_profile=meta_profile)


def add_to_sources(local_path=None, remote_path=None, parser=None):
    source_table = get_source_table()
    id = source_table.search_source(local=local_path, remote=remote_path)
    if id is None:
        id = source_table.add_source(local=local_path, remote=remote_path, parser=parser)
    return id


def load_from_remote(remote_name, owner=None):
    """ Loads the data from a remote repository.

    :param remote_name: The name of the dataset in the remote repository
    :param owner: (optional) The owner of the dataset. If nothing is provided, the current user
                  is used. For public datasets use 'public'.

    :return: A new GMQLDataset or a GDataframe
    """
    from .. import GMQLDataset
    pmg = get_python_manager()
    remote_manager = get_remote_manager()
    parser = remote_manager.get_dataset_schema(remote_name, owner)

    id = add_to_sources(remote_path=remote_name, parser=parser)
    index = pmg.read_dataset(str(id), parser.get_gmql_parser())
    remote_sources = [id]
    return GMQLDataset.GMQLDataset(index=index, location="remote", path_or_name=remote_name,
                                   remote_sources=remote_sources)


def load(path=None, name=None, owner=None, parser=None, all_load=False):
    # TODO: think if this method is useful or not...
    mode = get_mode()
    remote_manager = get_remote_manager()
    if mode == 'local':
        if isinstance(path, str) and (name is None):
            # we are given a local path
            return load_from_path(local_path=path, parser=parser, all_load=all_load)
        elif isinstance(name, str) and (path is None):
            local_path = TempFileManager.get_new_dataset_tmp_folder()
            remote_manager.download_dataset(dataset_name=name, local_path=local_path)
            return load_from_path(local_path=local_path, all_load=all_load)
        else:
            ValueError("You have to define path or name (mutually exclusive)")
    elif mode == 'remote':
        if isinstance(path, str) and (name is None):
            name = TempFileManager.get_unique_identifier()
            remote_manager.upload_dataset(dataset=path, dataset_name=name)
            return load_from_remote(remote_name=name)
        elif isinstance(name, str) and (path is None):
            return load_from_remote(remote_name=name, owner=owner)
        else:
            ValueError("You have to define path or name (mutually exclusive)")
    else:
        raise ValueError("Mode: {} unknown".format(mode))




import unittest
import gmql as gl
import shutil
import os
import logging

gl.set_progress(False)
gl.set_remote_address("http://gmql.eu/gmql-rest/")


class GMQLRemoteTester(unittest.TestCase):
    def setUp(self):
        log_fmt = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(level=logging.INFO, format=log_fmt)
        try:
            gl.login()
        except Exception:
            self.skipTest("Impossible to connect to the remote system! Remote testing will not be done!")
        self.rm = gl.get_remote_manager()
        self.remote_output_path = "./remote_out/"
        os.makedirs(self.remote_output_path)

    def tearDown(self):
        shutil.rmtree(self.remote_output_path)

    def gdataframe_equality(self, gd1, gd2):
        sample_mapping = {}

        r1, r2 = gd1.regs, gd2.regs
        m1, m2 = gd1.meta, gd2.meta

        # removing the samples having no regions
        m1 = m1.loc[set(m1.index).difference(set(m1.index).difference(set(r1.index)))]
        m2 = m2.loc[set(m2.index).difference(set(m2.index).difference(set(r2.index)))]

        coords_cols = ['chr', 'start', 'stop'] + (['strand'] if 'strand' in r1.columns else [])
        rest_cols = sorted(set(r1.columns).difference(set(coords_cols)))
        r1, r2 = r1.sort_values(coords_cols + rest_cols), r2.sort_values(coords_cols + rest_cols)

        for i, idx1 in enumerate(r1.index):
            sample_mapping[idx1] = r2.index[i]

        r1.index = r1.index.map(lambda x: sample_mapping[x])
        r1, r2 = r1.sort_index(), r2.sort_index()
        m1.index = m1.index.map(lambda x: sample_mapping.get(x))

        m1, m2 = m1.sort_index(0).sort_index(1), m2.sort_index(0).sort_index(1)

        self.assertTrue((m1.applymap(set) == m2.applymap(set)).all().all())
        self.assertTrue((r1.fillna(0) == r2.fillna(0)).all().all())

    def test_remote_select_1(self):
        querytext = """
        #SELECT 1
            RES = SELECT(region: (chr == chr2 OR chr == chr3) AND NOT(strand == + OR strand == -)
                AND start >= 130 AND stop <= 250) Example_Dataset_2;
            MATERIALIZE RES INTO select_1;
        """

        logging.info("Query: {}".format(querytext))
        logging.info("Executing REMOTE TEXTUAL query")
        respaths = self.rm.query(querytext, self.remote_output_path)
        dataset_name = respaths.iloc[0].dataset
        respaths = os.path.join(self.remote_output_path, dataset_name)
        logging.info("Deleting remote dataset {}".format(dataset_name))
        self.rm.delete_dataset(dataset_name)
        res_query = gl.load_from_path(respaths).materialize()

        logging.info("Executing LOCAL PYTHON query")
        d = gl.get_example_dataset("Example_Dataset_2")
        d = d.select(region_predicate=(d.chr.isin(['chr2', 'chr3'])) & (~d.strand.isin(['+', '-'])) &
                                      (d.start >= 130) & (d.stop <= 250))
        res_local = d.materialize()
        self.gdataframe_equality(res_query, res_local)

        logging.info("Executing REMOTE PYTHON query")
        gl.set_mode("remote")
        res_remote = d.materialize()
        self.gdataframe_equality(res_local, res_remote)


if __name__ == '__main__':
    unittest.main()



# class MetadataParser:
#     """ Abstract Parser of Metadata files
#
#     """
#     def parse_metadata(self, path):
#         """ Parse a metadata file into a list of tuple
#         [('attr_name', attr_value'), ...]
#
#         :param path: path of the metadata file
#         :return: dict
#         """
#         pass
#
#
# class GenericMetaParser(MetadataParser):
#     def __init__(self, delimiter="\t"):
#         self.delimiter = delimiter
#
#     def parse_metadata(self, path):
#         fo = open(path)
#         lines = fo.read().splitlines()
#         fo.close()
#         return list(map(self._parse_line_metadata, lines))
#
#     def _parse_line_metadata(self, line):
#         elems = line.split(self.delimiter)
#         return elems[0], elems[1]



class Configuration:
    """ Class containing all the information regarding the system environment and the Spark environment
    """

    def __init__(self):
        self.app_name = "PyGMQL"
        self.master = "local[*]"
        self.sc = None
        self._properties = dict()
        self._system = dict()

    def get_spark_confs(self):
        return self._properties

    def get_system_confs(self):
        return self._system

    def set_spark_context(self, sc):
        self.sc = sc

    def set_app_name(self, name):
        """ Sets the name of the application in spark,
        By default it is called "gmql_api"

        :param name: string
        :return: None
        """
        self.app_name = name

    def set_master(self, master):
        """ Set the master of the spark cluster
        By default it is "local[*]"

        :param master: string
        :return: None
        """
        self.master = master

    def set_spark_conf(self, key=None, value=None, d=None):
        """ Sets a spark property as a ('key', 'value') pair of using a dictionary
        {'key': 'value', ...}

        :param key: string
        :param value: string
        :param d: dictionary
        :return: None
        """
        if isinstance(d, dict):
            self._properties.update(d)
        elif isinstance(key, str) and isinstance(value, str):
            self._properties[key] = value
        else:
            raise TypeError("key, value must be strings")

    def set_system_conf(self, key=None, value=None, d=None):
        """ Sets a java system property as a ('key', 'value') pair of using a dictionary
        {'key': 'value', ...}

        :param key: string
        :param value: string
        :param d: dictionary
        :return: None
        """
        if isinstance(d, dict):
            self._system.update(d)
        elif isinstance(key, str) and isinstance(value, str):
            self._system[key] = value
        else:
            raise TypeError("key, value must be strings")


import unittest
import gmql as gl
from gmql.dataset.GDataframe import GDataframe
from gmql.dataset.GMQLDataset import GMQLDataset
from pkg_resources import resource_filename
import os
from glob import glob
import shutil

gl.set_progress(False)


class GMQLQueryTester(unittest.TestCase):

    def setUp(self):
        self.query_results_path = resource_filename("gmql", os.path.join("resources", "example_queries_results"))

    def get_query_results(self, queryname):
        datapath = glob(os.path.join(self.query_results_path, "*{}".format(queryname)))[0]
        return gl.load_from_path(datapath).materialize()

    def gdataframe_equality(self, gd1, gd2):
        sample_mapping = {}

        r1, r2 = gd1.regs, gd2.regs
        m1, m2 = gd1.meta, gd2.meta

        # removing the samples having no regions
        m1 = m1.loc[set(m1.index).difference(set(m1.index).difference(set(r1.index)))]
        m2 = m2.loc[set(m2.index).difference(set(m2.index).difference(set(r2.index)))]

        coords_cols = ['chr', 'start', 'stop'] + (['strand'] if 'strand' in r1.columns else [])
        rest_cols = sorted(set(r1.columns).difference(set(coords_cols)))
        r1, r2 = r1.sort_values(coords_cols + rest_cols), r2.sort_values(coords_cols + rest_cols)

        for i, idx1 in enumerate(r1.index):
            sample_mapping[idx1] = r2.index[i]

        r1.index = r1.index.map(lambda x: sample_mapping[x])
        r1, r2 = r1.sort_index(), r2.sort_index()
        m1.index = m1.index.map(lambda x: sample_mapping.get(x))

        m1, m2 = m1.sort_index(0).sort_index(1), m2.sort_index(0).sort_index(1)

        self.assertTrue((m1.applymap(set) == m2.applymap(set)).all().all())
        self.assertTrue((r1.fillna(0) == r2.fillna(0)).all().all())

    def test_load_example_datasets(self):
        self.assertIsInstance(gl.get_example_dataset(name="Example_Dataset_1"), GMQLDataset)
        self.assertIsInstance(gl.get_example_dataset(name="Example_Dataset_2"), GMQLDataset)

        self.assertIsInstance(gl.get_example_dataset(name="Example_Dataset_1", load=True), GDataframe)
        self.assertIsInstance(gl.get_example_dataset(name="Example_Dataset_2", load=True), GDataframe)

    def test_select_1(self):
        """ #SELECT 1
            RES = SELECT(region: (chr == chr2 OR chr == chr3) AND NOT(strand == + OR strand == -)
                AND start >= 130 AND stop <= 250) Example_Dataset_2;
            MATERIALIZE RES INTO select_1;
        """

        res = gl.get_example_dataset("Example_Dataset_2")
        res = res.select(region_predicate=(res.chr.isin(['chr2', 'chr3'])) & (~res.strand.isin(['+', '-'])) &
                                          (res.start >= 130) & (res.stop <= 250))
        res = res.materialize()
        self.gdataframe_equality(res, self.get_query_results("select_1"))

    def test_project_1(self):
        """ # project_1
            D = SELECT() Example_Dataset_1;
            RES = PROJECT(region_update: length AS right - left) D;
            MATERIALIZE RES INTO project_1;
        """

        res = gl.get_example_dataset("Example_Dataset_1")
        res = res.project(new_field_dict={'length': res.right - res.left})
        res = res.materialize()
        self.gdataframe_equality(res, self.get_query_results("project_1"))

    def test_extend_2(self):
        """ # extend_2
            D = SELECT() Example_Dataset_1;
            RES = EXTEND(region_count AS COUNT(), min_pvalue AS MIN(pvalue)) D;
            MATERIALIZE RES INTO extend_2;
        """

        res = gl.get_example_dataset("Example_Dataset_1")
        res = res.extend({'region_count': gl.COUNT(), 'min_pvalue': gl.MIN("pvalue")})
        res = res.materialize()
        self.gdataframe_equality(res, self.get_query_results("extend_2"))

    def test_order_1(self):
        """ # order_1
            D = SELECT(region: chr == chr1) Example_Dataset_1;
            D1 = EXTEND(Region_count AS COUNT()) D;
            RES = ORDER(Region_count DESC; meta_top: 2) D1;
            MATERIALIZE RES INTO order_1;
        """

        res = gl.get_example_dataset("Example_Dataset_1")
        res = res.reg_select(res.chr == 'chr1')
        res = res.extend({'Region_count': gl.COUNT()})
        res = res.order(meta=['Region_count'], meta_ascending=[False], meta_top="top", meta_k=2)
        res = res.materialize()
        self.gdataframe_equality(res, self.get_query_results("order_1"))

    def test_group_1(self):
        """ # group_1
            D = SELECT(region: chr == chr2) Example_Dataset_2;
            RES = GROUP(controlId; meta_aggregates: max_cell_tier AS MAX(cell_tier)) D;
            MATERIALIZE RES INTO group_1;
        """

        res = gl.get_example_dataset("Example_Dataset_2")
        res = res.reg_select(res.chr == 'chr2')
        res = res.group(meta=['controlId'], meta_aggregates={'max_cell_tier': gl.MAX("cell_tier")})
        res = res.materialize()
        self.gdataframe_equality(res, self.get_query_results("group_1"))

    def test_merge_1(self):
        """ # merge_1
            D = SELECT(region: chr == chr1) Example_Dataset_1;
            RES = MERGE() D;
            MATERIALIZE RES INTO merge_1;
        """

        res = gl.get_example_dataset("Example_Dataset_1")
        res = res.reg_select(res.chr == 'chr1')
        res = res.merge()
        res = res.materialize()
        self.gdataframe_equality(res, self.get_query_results("merge_1"))

    def test_difference_1(self):
        """ # difference_1
            D1 = SELECT(region: chr == chr2) Example_Dataset_1;
            D2 = SELECT(cell_karyotype == "cancer"; region: chr == chr2) Example_Dataset_2;
            RES = DIFFERENCE() D1 D2;
            MATERIALIZE RES INTO difference_1;
        """

        d1 = gl.get_example_dataset("Example_Dataset_1")
        d1 = d1.reg_select(d1.chr == 'chr2')

        d2 = gl.get_example_dataset("Example_Dataset_2")
        d2 = d2.select(meta_predicate=d2['cell_karyotype'] == 'cancer', region_predicate=d2.chr == 'chr2')

        res = d1.difference(d2)
        res = res.materialize()
        self.gdataframe_equality(res, self.get_query_results("difference_1"))

    def test_map_1(self):
        """ # map_1
            D1 = SELECT(region: chr == chr2) Example_Dataset_1;
            D2 = SELECT(region: chr == chr2) Example_Dataset_2;
            RES = MAP(avg_score AS AVG(score)) D1 D2;
            MATERIALIZE RES INTO map_1;
        """

        d1 = gl.get_example_dataset("Example_Dataset_1")
        d1 = d1.reg_select(d1.chr == 'chr2')

        d2 = gl.get_example_dataset("Example_Dataset_2")
        d2 = d2.reg_select(d2.chr == 'chr2')

        res = d1.map(d2, new_reg_fields={'avg_score': gl.AVG("score")}, refName="D1", expName="D2")
        res = res.materialize("./tmp")
        shutil.rmtree("./tmp")
        self.gdataframe_equality(res, self.get_query_results("map_1"))

    def test_join_1(self):
        """ # join_1
            D1 = SELECT(region: chr == chr2) Example_Dataset_1;
            D2 = SELECT(region: chr == chr2) Example_Dataset_2;
            RES = JOIN(MD(1), DGE(20); output: RIGHT; joinby: cell_karyotype) D1 D2;
            MATERIALIZE RES INTO join_1;
        """

        d1 = gl.get_example_dataset("Example_Dataset_1")
        d1 = d1.reg_select(d1.chr == 'chr2')

        d2 = gl.get_example_dataset("Example_Dataset_2")
        d2 = d2.reg_select(d2.chr == 'chr2')

        res = d1.join(d2, [gl.MD(1), gl.DGE(20)], output="RIGHT", joinBy=['cell_karyotype'], refName="D1", expName="D2")
        res = res.materialize("./tmp")
        shutil.rmtree("./tmp")
        self.gdataframe_equality(res, self.get_query_results("join_1"))

    def test_cover_2(self):
        """ # cover_2
            D = SELECT(region: chr == chr2) Example_Dataset_2;
            RES = COVER(2, 3; groupby: cell; aggregate: min_pvalue AS MIN(pvalue)) D;
            MATERIALIZE RES INTO cover_2;
        """

        d = gl.get_example_dataset("Example_Dataset_2")
        d = d.reg_select(d.chr == 'chr2')
        d = d.normal_cover(2, 3, groupBy=['cell'], new_reg_fields={'min_pvalue': gl.MIN("pvalue")})
        d = d.materialize()
        self.gdataframe_equality(d, self.get_query_results("cover_2"))


    # @staticmethod
    # def test_complex_query():
    #     d = gl.load_from_path("../../data/narrow_sample")
    #     d = d.reg_select(d.pvalue >= 0.05)
    #     d = d[d['experiment_target'] == 'ARID3A']
    #     d = d.reg_project(new_field_dict={
    #         'newField': (d.start + d.pvalue) / 2
    #     })
    #     d = d.join(d, genometric_predicate=[gl.DLE(100000)])
    #     d = d.map(d)
    #     d = d.cover(minAcc=1, maxAcc="ANY")
    #     d = d.materialize()


if __name__ == '__main__':
    unittest.main()



from .FileManagment import TempFileManager, initialize_user_folder, get_resources_dir
from .configuration import Configuration
import os
import sys


__version__ = None
__progress_bar = True
__metadata_profiling = False
__remote_address = None
__mode = "local"
__master = "local[*]"
__gcloud_token = None
__folders = None
__init_configs_local = {
    "spark.serializer": 'org.apache.spark.serializer.KryoSerializer',
    'spark.executor.memory': '6g',
    'spark.driver.memory': '8g',
    'spark.kryoserializer.buffer.max': '1g',
    'spark.driver.maxResultSize': '5g',
    'spark.driver.host': 'localhost',
    'spark.local.dir': '/tmp'
}
__regions_batch_size = 20000
__init_configs_spark = {}
__configuration = None
__java_options = []  # ['-Xmx8192m']


def get_configuration():
    """ Returns the configurations of the current PyGMQL instance

    :return: a Configuration object
    """
    global __configuration
    return __configuration


def set_spark_configs(d):
    """ Set Spark configurations to be used during the spark-submit. Works only when the master is different from local.

    :param d: a dictionary of {key: values}
    :return: None
    """
    global __init_configs_spark
    __init_configs_spark.update(d)


def get_spark_configs():
    global __init_configs_spark
    return __init_configs_spark


def set_regions_batch_size(batch: int):
    """ Set the number of regions to be loaded for each batch at the end of the materialize operation.
    This setting is very low level and does not affect the result of the query, but changing this parameter
    could speed up the loading of the results in Python in certain contexts.

    :param batch: Number of regions.
    :return: None
    """
    global __regions_batch_size
    __regions_batch_size = batch


def get_regions_batch_size():
    global __regions_batch_size
    return __regions_batch_size


def set_configuration(conf):
    global __configuration
    if not isinstance(conf, Configuration):
        raise TypeError("Configuration expected. {} was found".format(type(conf)))
    __configuration = conf


def set_master(master: str):
    """ Set the master of the PyGMQL instance. It accepts any master configuration available in Spark.

    :param master: master configuration
    :return: None
    """
    global __master
    __master = master


def get_master():
    global __master
    return __master


def set_local_java_options(options: list):
    """ When the mode is set to local, this function can be used to add JVM specific options before starting the
    backend. It accepts any Java option.

    :param options: list of string, one for each Java option
    :return: None
    """
    global __java_options
    __java_options = options


def get_local_java_options():
    global __java_options
    return __java_options


def set_mode(how):
    """ Sets the behavior of the API

    :param how: if 'remote' all the execution is performed on the remote server; if 'local' all
           it is executed locally. Default = 'local'
    :return: None
    """
    global __mode
    if how == "local":
        __mode = how
    elif how == "remote":
        __mode = how
    else:
        raise ValueError("how must be 'local' or 'remote'")


def get_mode():
    global __mode
    return __mode


def get_version():
    version_file_name = os.path.join(get_resources_dir(),  "version")
    with open(version_file_name, "r") as f_ver:
        version = f_ver.read().strip()
    return version


def set_progress(how):
    """ Enables or disables the progress bars for the loading, writing and downloading
    of datasets

    :param how: True if you want the progress bar, False otherwise
    :return: None

    Example::

        import gmql as gl

        gl.set_progress(True)   # abilitates progress bars
        # ....do something...
        gl.set_progress(False)  # removes progress bars
        # ....do something...
    """
    global __progress_bar
    if isinstance(how, bool):
        __progress_bar = how
    else:
        raise ValueError(
            "how must be a boolean. {} was found".format(type(how)))


def is_progress_enabled():
    global __progress_bar
    return __progress_bar


def set_meta_profiling(how):
    """ Enables or disables the profiling of metadata at the loading of a GMQLDataset

    :param how: True if you want to analyze the metadata when a GMQLDataset is created
                by a load_from_*. False otherwise. (Default=True)
    :return: None
    """
    global __metadata_profiling
    if isinstance(how, bool):
        __metadata_profiling = how
    else:
        raise TypeError("how must be boolean. {} was provided".format(type(how)))


def is_metaprofiling_enabled():
    global __metadata_profiling
    return __metadata_profiling


def set_remote_address(address):
    """ Enables the user to set the address of the GMQL remote service

    :param address: a string representing the URL of GMQL remote service
    :return: None
    """
    global __remote_address
    __remote_address = address


def get_remote_address():
    global __remote_address
    return __remote_address


def get_folders():
    global __folders
    return __folders


def initialize_configuration():
    global __configuration, __init_configs_local
    configs = Configuration()
    configs.set_spark_conf(d=__init_configs_local)
    __configuration = configs


def init_settings():
    global __version__, __folders, __configuration
    __version__ = get_version()
    initialize_user_folder()
    __folders = TempFileManager.initialize_tmp_folders()
    initialize_configuration()
    __configuration.set_spark_conf("spark.local.dir", __folders['spark'])
    if sys.platform.startswith("win32"):
        # if we are on windows set the hadoop home to winutils.exe
        hadoop_folder_fn = os.path.join(get_resources_dir(), "hadoop")
        __configuration.set_system_conf("hadoop.home.dir", hadoop_folder_fn)



import os
from pkg_resources import resource_filename


user_path = os.path.expanduser("~")
USER_FOLDER_NAME = ".pygmql"

user_folder = os.path.join(user_path, USER_FOLDER_NAME)
resource_folder = resource_filename("gmql", "resources")


def initialize_user_folder():
    if not os.path.isdir(get_user_dir()):
        os.makedirs(get_user_dir())


def get_user_dir():
    return user_folder


def get_resources_dir():
    return resource_folder



from .loaders.Loader import load_from_path
from ..FileManagment import get_resources_dir
import os


def get_example_dataset(name="Example_Dataset_1", load=False):
    data_path = os.path.join(get_resources_dir(), "example_datasets", name)
    res = load_from_path(data_path)
    if load:
        res = res.materialize()
    return res



from py4j.java_collections import ListConverter
from .managers import get_gateway, get_python_manager


def to_java_list(l):
    gateway = get_gateway()
    ListConverter().convert(l, gateway)


def none():
    return get_python_manager().getNone()


def Some(thing):
    return get_python_manager().getSome(thing)



import pandas as pd

LOCAL = "local"
DELETE_LOCAL = "delete_local"
REMOTE = "remote"
DELETE_REMOTE = "delete_remote"
PARSER = "parser"


class SourcesTable:
    """ Handles the location of the datasets during the program execution.
    Manages both local and remote datasets. It is used when a mode change is
    detected to send or download data sources to/from the remote server.
    """
    def __init__(self):
        self.table = pd.DataFrame(columns=[LOCAL, DELETE_LOCAL,  REMOTE, DELETE_REMOTE, PARSER])
        self.id_count = 0

    def add_source(self, local=None, delete_local=False, remote=None, delete_remote=False, parser=None):
        self.table = self.table.append({LOCAL: local,
                                        DELETE_LOCAL: delete_local,
                                        REMOTE: remote,
                                        DELETE_REMOTE: delete_remote,
                                        PARSER: parser}, ignore_index=True)
        result = self.id_count
        self.id_count += 1
        return result

    def get_source(self, id):
        res = self.table.loc[id].to_dict()
        return res

    def modify_source(self, id, local=None, delete_local=None, remote=None, delete_remote=None):
        if local is not None:
            self.table.loc[id][LOCAL] = local
        if delete_local is not None:
            self.table.loc[id][DELETE_LOCAL] = delete_local
        if remote is not None:
            self.table.loc[id][REMOTE] = remote
        if delete_remote is not None:
            self.table.loc[id][DELETE_REMOTE] = delete_remote

    def search_source(self, local=None, remote=None):
        if local is not None:
            ids = self.table[self.table[LOCAL] == local].index.tolist()
        elif remote is not None:
            ids = self.table[self.table[REMOTE] == remote].index.tolist()
        else:
            raise ValueError("local or remote must be given")
        return ids[0] if len(ids) > 0 else None

    def get_deletable(self, where="remote"):
        if where == 'remote':
            names = self.table.loc[self.table[DELETE_REMOTE], REMOTE].tolist()
        elif where == 'local':
            names = self.table.loc[self.table[DELETE_LOCAL], LOCAL].tolist()
        else:
            raise ValueError("where must be 'remote' or 'local'")
        return names



# import os
# from . import SCHEMA_FILE


# def load_reg_from_path(path, parser=None):
#     if parser is None:
#         # get the parser for the dataset
#         parser = RegionParser.from_schema_file(path)
#     # we need to take only the files of the regions, so only the files that does NOT end with '.meta'
#     all_files = set(glob(pathname=path + '/*'))
#     meta_files = set(glob(pathname=path + '/*.gdm.meta'))
#
#     only_region_files = all_files - meta_files
#     # print(only_region_files)
#     logger = logging.getLogger()
#     logger.debug("Loading region data from path {}".format(path))
#
#     n_files = len(only_region_files)
#     dfs = []
#     for file in tqdm.tqdm(only_region_files, total=n_files, disable=not is_progress_enabled()):
#         if file.endswith("schema") or file.endswith("_SUCCESS") or \
#            file.endswith(".xml"):
#             continue
#         abs_path = os.path.abspath(file)
#         key = generateNameKey(abs_path)
#
#         df = parser.parse_regions(abs_path)
#         df.index = np.repeat(key, len(df))
#         dfs.append(df)
#     if len(dfs) > 0:
#         result = pd.concat(objs=dfs, copy=False)
#     else:
#         result = pd.DataFrame(columns=parser.get_ordered_attributes())
#     result.index.name = "id_sample"
#     # ordering the columns
#     new_reg_fixed_fields = reg_fixed_fileds.copy()
#     if parser.strandPos is None:
#         new_reg_fixed_fields.remove("strand")
#     result = result[new_reg_fixed_fields + [c for c in result.columns if c not in new_reg_fixed_fields]]
#     # sorting index
#     result = result.sort_index()
#     return result


# def getHdfsSchema(path):
#     path = getGCSchemaPath(path)
#     client = InsecureClient('http://ip-172-31-37-168.us-east-2.compute.internal:50070', 'hadoop')
#     with client.read(path[5:]) as reader:
#         xml = reader.read().decode("utf-8")
#     tree = ET.ElementTree(ET.fromstring(xml))
#     return tree
#
#
# def getGCSchemaPath(path):
#     if path.endswith("/"):
#         return path + SCHEMA_FILE
#     else:
#         return path + "/" + SCHEMA_FILE
#
#
# def getGCSchema(path):
#     fs = gcsfs.GCSFileSystem(token=get_gcloud_token())
#     xml = fs.read_block(getGCSchemaPath(path), 0, None).decode("utf-8")
#     tree = ET.ElementTree(ET.fromstring(xml))
#     return tree
#
#
# def getLocalSchema(path):
#     schema_file = get_schema_path(path)
#     tree = ET.parse(schema_file)
#     return tree
#
#
# def __get_parser_old(path):
#     if path.startswith("gs://"):
#         tree = getGCSchema(path)
#     elif path.startswith("hdfs://"):
#         tree = getHdfsSchema(path)
#     else:
#         tree = getLocalSchema(path)
#     gmqlSchema = tree.getroot().getchildren()[0]
#     parser_type = gmqlSchema.get('type')
#     coordinate_system = gmqlSchema.get("coordinate_system")
#     if coordinate_system is None:
#         coordinate_system = COORDS_DEFAULT
#     field_nodes = gmqlSchema.getchildren()
#
#     i = 0
#     chrPos, startPos, stopPos, strandPos, otherPos = None, None, None, None, None
#     otherPos = []
#
#     if parser_type == GTF:
#         chrPos = 0      # seqname
#         startPos = 3    # start
#         stopPos = 4     # end
#         strandPos = 6   # strand
#         otherPos = [(1, 'source', 'string'), (2, 'feature', 'string'),
#                     (5, 'score', 'float'), (7, 'frame', 'string')]
#
#         for field in field_nodes:
#             name = list(field.itertext())[0]#.lower()
#             type = field.get('type').lower()
#             if name not in {'seqname', 'start', 'end', 'strand',
#                             'source', 'feature', 'score', 'frame'}:
#                 otherPos.append((i, name, type))
#             i += 1
#
#     else:
#         for field in field_nodes:
#             name = list(field.itertext())[0]#.lower()
#             type = field.get('type').lower()
#
#             if name in chr_aliases and chrPos is None:
#                 chrPos = i
#             elif name in start_aliases and startPos is None:
#                 startPos = i
#             elif name in stop_aliases and stopPos is None:
#                 stopPos = i
#             elif name in strand_aliases and strandPos is None:
#                 strandPos = i
#             else:   # other positions
#                 otherPos.append((i, name, type))
#             i += 1
#     if len(otherPos) == 0:
#         otherPos = None
#     return RegionParser(chrPos, startPos, stopPos, strandPos,
#                         otherPos, coordinate_system=coordinate_system,
#                         schema_format=parser_type)






from ...managers import get_python_manager, get_gateway
from ...scala_wrapper import none, Some
from . import coordinate_systems, get_parsing_function, null_values, GTF
import numpy as np
import pandas as pd
import re


class RegionParser:
    def __init__(self, gmql_parser=None, chrPos=None, startPos=None, stopPos=None,
                 strandPos=None, otherPos=None, delimiter="\t",
                 coordinate_system='0-based', schema_format="del",
                 parser_name="parser"):
        """ Creates a custom region dataset

        :param chrPos: position of the chromosome column
        :param startPos: position of the start column
        :param stopPos: position of the stop column
        :param strandPos: (optional) position of the strand column. Default is None
        :param otherPos: (optional) list of tuples of the type [(pos, attr_name, typeFun), ...]. Default is None
        :param delimiter: (optional) delimiter of the columns of the file. Default "\t"
        :param coordinate_system: can be {'0-based', '1-based', 'default'}. Default is '0-based'
        :param schema_format: (optional) type of file. Can be {'tab', 'gtf', 'vcf', 'del'}. Default is 'del'
        :param parser_name: (optional) name of the parser. Default is 'parser'
        """

        if isinstance(parser_name, str):
            self.parser_name = parser_name
        else:
            raise TypeError("Parser name must be a string")

        if gmql_parser is None:
            if not isinstance(delimiter, str):
                raise ValueError("delimiter must be a string")
            if not (isinstance(chrPos, int) and chrPos >= 0):
                raise ValueError("Chromosome position must be >=0")
            if not (isinstance(startPos, int) and startPos >= 0):
                raise ValueError("Start position must be >=0")
            if not (isinstance(stopPos, int) and stopPos >= 0):
                raise ValueError("Stop position must be >=0")
            if isinstance(strandPos, int) and strandPos >= 0:
                strandGmql = Some(strandPos)
            elif strandPos is None:
                strandGmql = none()
            else:
                raise ValueError("Strand position must be >= 0")
            if not isinstance(coordinate_system, str):
                raise TypeError("Coordinate system must be a string")
            if coordinate_system not in coordinate_systems:
                raise ValueError("{} is not a valid coordinate system".format(coordinate_system))
            if not isinstance(schema_format, str):
                raise TypeError("Schema Format must be a string")
            if isinstance(otherPos, list):
                otherPosGmql = Some(convert_to_gmql(otherPos))
            else:
                otherPosGmql = none()

            pmg = get_python_manager()

            self.gmql_parser = pmg.buildParser(delimiter, chrPos, startPos,
                                               stopPos, strandGmql, otherPosGmql,
                                               schema_format, coordinate_system)
        else:
            self.gmql_parser = gmql_parser

    @staticmethod
    def from_schema_file(schema_file):
        pmg = get_python_manager()
        gmql_parser = pmg.getParserFromPath(schema_file)
        return RegionParser(gmql_parser)

    @property
    def delimiter(self):
        return self.gmql_parser.delimiter()

    @property
    def chrPos(self):
        return self.gmql_parser.chrPos()

    @property
    def startPos(self):
        return self.gmql_parser.startPos()

    @property
    def stopPos(self):
        return self.gmql_parser.stopPos()

    @property
    def strandPos(self):
        if self.gmql_parser.strandPos().isDefined():
            return self.gmql_parser.strandPos().get()
        else:
            return None

    @property
    def otherPos(self):
        res = []
        if self.gmql_parser.otherPos().isDefined():
            for poss, sch in zip(self.gmql_parser.otherPos().get(), self.gmql_parser.getSchema()):
                pos = poss._1()
                attr_name = sch._1()
                typeFun = get_parsing_function(poss._2().toString().lower())
                res.append((pos, attr_name, typeFun))
        return res

    def get_coordinates_system(self):
        return self.gmql_parser.coordinateSystem().toString()

    def get_parser_type(self):
        return self.gmql_parser.parsingType().toString()

    def get_gmql_parser(self):
        """ Gets the Scala implementation of the parser

        :return: a Java Object
        """
        return self.gmql_parser

    @staticmethod
    def parse_strand(strand):
        """ Defines how to parse the strand column

        :param strand: a string representing the strand
        :return: the parsed result
        """
        if strand in ['+', '-', '*']:
            return strand
        else:
            return '*'

    def parse_regions(self, path):
        """ Given a file path, it loads it into memory as a Pandas dataframe

        :param path: file path
        :return: a Pandas Dataframe
        """
        if self.get_parser_type().lower() == GTF.lower():
            res = self._parse_gtf_regions(path)
        else:
            res = self._parse_tab_regions(path)
        return res

    def _parse_tab_regions(self, path):
        types = self.get_name_type_dict()
        if "strand" in types.keys():
            types.pop("strand")
        fo = open(path)
        df = pd.read_csv(filepath_or_buffer=fo, na_values=null_values,
                         header=None,
                         names=self.get_ordered_attributes(),
                         dtype=types,
                         sep=self.delimiter,
                         converters={'strand': self.parse_strand})
        fo.close()
        return df

    def _parse_gtf_regions(self, path):
        def split_attributes(attrs):
            res = {}
            attr_splits = re.split(";\s*", attrs)
            for a in attr_splits:
                if len(a) > 0:
                    splits = re.split("\"\s*", a)
                    #             print(splits)
                    attr_name = splits[0].strip()
                    attr_value = splits[1].strip()
                    res[attr_name] = attr_value
            return res
        actual_attributes = self.get_ordered_attributes()[:8] + ['attributes']
        types = self.get_name_type_dict()
        types.pop("strand")
        fo = open(path)
        df = pd.read_csv(filepath_or_buffer=fo, sep=self.delimiter,
                         na_values=null_values,
                         names=actual_attributes,
                         dtype=types,
                         converters={'strand': self.parse_strand})
        fo.close()
        df = pd.concat([df, pd.DataFrame(df.attributes.map(split_attributes).tolist())], axis=1)\
            .drop("attributes", axis=1)
        return df

    def get_attributes(self):
        """ Returns the unordered list of attributes

        :return: list of strings
        """
        attr = ['chr', 'start', 'stop']
        if self.strandPos is not None:
            attr.append('strand')
        for i, o in enumerate(self.otherPos):
            attr.append(o[1])

        return attr

    def get_ordered_attributes(self):
        """ Returns the ordered list of attributes

        :return: list of strings
        """
        attrs = self.get_attributes()
        attr_arr = np.array(attrs)
        poss = [self.chrPos, self.startPos, self.stopPos]
        if self.strandPos is not None:
            poss.append(self.strandPos)
        for o in self.otherPos:
            poss.append(o[0])
        idx_sort = np.array(poss).argsort()
        return attr_arr[idx_sort].tolist()

    def get_types(self):
        """ Returns the unordered list of data types

        :return: list of data types
        """
        types = [str, int, int]
        if self.strandPos is not None:
            types.append(str)
        for o in self.otherPos:
            types.append(o[2])
        return types

    def get_name_type_dict(self):
        """ Returns a dictionary of the type
        {'column_name': data_type, ...}

        :return: dict
        """
        attrs = self.get_attributes()
        types = self.get_types()
        d = dict()
        for i, a in enumerate(attrs):
            d[a] = types[i]

        return d

    def get_ordered_types(self):
        """ Returns the ordered list of data types

        :return: list of data types
        """
        types = self.get_types()
        types_arr = np.array(types)
        poss = [self.chrPos, self.startPos, self.stopPos]
        if self.strandPos is not None:
            poss.append(self.strandPos)
        if self.otherPos:
            for o in self.otherPos:
                poss.append(o[0])
        idx_sort = np.array(poss).argsort()
        return types_arr[idx_sort].tolist()


def convert_otherPos(otherPos):
    # print(otherPos)
    return list(map(_to_parsing_function, otherPos))


def _to_parsing_function(tpos):
    if len(tpos) != 3:
        raise ValueError("Position tuple has wrong number of parameters")
    else:
        if isinstance(tpos[2], str):
            fun = get_parsing_function(tpos[2].lower())
        else:
            raise TypeError("Type of region field must be a string")
        if isinstance(tpos[0], int) and isinstance(tpos[1], str):
            return tpos[0], tpos[1], fun
        else:
            raise TypeError("Position must be integer and name of field must be a string")


def convert_to_gmql(otherPos):
    otherPosJavaList = get_gateway().jvm.java.util.ArrayList()
    for tpos in otherPos:
        posJavaList = get_gateway().jvm.java.util.ArrayList()
        if len(tpos) != 3:
            raise ValueError("Position tuple has wrong number of parameters")
        else:
            if isinstance(tpos[0], int) and tpos[0] >= 0:
                posJavaList.append(str(tpos[0]))
            else:
                raise ValueError("Position must be >= 0")
            if isinstance(tpos[1], str):
                posJavaList.append(tpos[1])
            else:
                raise TypeError("Name of the field must be a string")
            if isinstance(tpos[2], str):
                get_parsing_function(tpos[2])
                posJavaList.append(tpos[2])
        otherPosJavaList.append(posJavaList)
    return otherPosJavaList



from .FileManagment.DependencyManager import DependencyManager
from .FileManagment import get_user_dir
from .RemoteConnection.SessionManager import load_sessions, store_sessions
from .FileManagment import TempFileManager
from .settings import get_remote_address, get_configuration, get_master, get_local_java_options,  get_spark_configs
from .configuration import Configuration
import py4j
from subprocess import Popen, PIPE
from py4j.java_gateway import JavaGateway, launch_gateway, GatewayParameters, \
    OutputConsumer, ProcessConsumer, quiet_close
import os
import time
import atexit
import signal
import warnings
import logging
import re
from findspark import find

__remote_manager = None
__session_manager = None
__dependency_manager = None
__source_table = None
__gateway = None
__pythonManager = None
__gmql_jar_path = None
__py4j_path = None


def start():
    global __pythonManager, __gateway, __dependency_manager, __gmql_jar_path, __py4j_path
    logger = logging.getLogger()
    master = get_master()

    if master.lower().startswith('local'):
        logger.debug("Starting LOCAL backend (master: {})".format(master.lower()))
        java_home = os.environ.get("JAVA_HOME")
        if java_home is None:
            raise SystemError("The environment variable JAVA_HOME is not set")
        java_path = os.path.join(java_home, "bin", "java")
        _port = launch_gateway(classpath=__gmql_jar_path, die_on_exit=True,
                               java_path=java_path, javaopts=get_local_java_options(),
                               jarpath=__py4j_path)
        __gateway = JavaGateway(gateway_parameters=GatewayParameters(port=_port,
                                                                     auto_convert=True))
        python_api_package = get_python_api_package(__gateway)
        __pythonManager = start_gmql_manager(python_api_package)

        conf = get_configuration()
        conf.set_master(master.lower())
        _set_spark_configuration(conf)
        _set_system_configuration(conf)
    else:
        # use spark-submit
        logger.debug("Submitting backend to {}".format(master))
        master = re.sub("^spark_", "", master.lower())
        configs = get_spark_configs()
        spark_location = find()
        logger.debug("Found spark at location: {}".format(spark_location))
        command = [os.path.join(spark_location, 'bin', 'spark-submit'), '--master', master, '--deploy-mode', "client"]

        for cname, c in configs.items():
            command.extend(['--conf', '{}={}'.format(cname, c)])

        command.append(__gmql_jar_path)

        stderr = open(os.devnull, "w")
        proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=stderr)

        while True:
            try:
                _port = int(proc.stdout.readline())
                break
            except ValueError:
                pass

        logger.debug("Backend listening at port {}".format(_port))
        redirect_stdout = open(os.devnull, "w")
        OutputConsumer(redirect_stdout, proc.stdout, daemon=True).start()
        ProcessConsumer(proc, [redirect_stdout], daemon=True).start()
        quiet_close(stderr)

        __gateway = JavaGateway(gateway_parameters=GatewayParameters(port=_port,
                                                                     auto_convert=True))
        pm = __gateway.entry_point.getPythonManager()
        pm.startEngine()
        __pythonManager = pm


def _set_spark_configuration(conf):
    if not isinstance(conf, Configuration):
        raise TypeError("Configuration is required. {} was passed".format(type(conf)))
    pmg = get_python_manager()
    pmg.setSparkConfiguration(conf.app_name,
                              conf.master,
                              conf.get_spark_confs())


def _set_system_configuration(conf):
    if not isinstance(conf, Configuration):
        raise TypeError("Configuration is required. {} was passed".format(type(conf)))
    pmg = get_python_manager()
    pmg.setSystemConfiguration(conf.get_system_confs())


def __check_py4j_backend():
    py4j_version = py4j.__version__
    py4j_backend_jar = os.path.join(get_user_dir(), "py4j-{}.jar".format(py4j_version))
    if not os.path.isfile(py4j_backend_jar):
        py4j_location = DependencyManager.find_package(
                            repo="https://oss.sonatype.org/content/repositories/releases/",
                            repo_name="releases",
                            groupId="net.sf.py4j",
                            artifactId="py4j",
                            version=py4j_version,
                        )
        DependencyManager.download_from_location(py4j_location, py4j_backend_jar)
    return py4j_backend_jar


def set_backend_path(path):
    """ Manually set the scala backend of the library

    :param path: location of the jar file
    :return: None
    """
    global __gmql_jar_path
    if not is_backend_on():
        __gmql_jar_path = path


def set_py4j_path(path):
    """ Manually set the py4j backend of the library

    :param path: location of the jar file
    :return: None
    """
    global __py4j_path
    if not is_backend_on():
        __py4j_path = path


def stop():
    global __gateway, __session_manager, __remote_manager, __source_table
    # storing the session
    store_sessions(__session_manager.sessions)
    # killing the gateway
    if __gateway is not None:
        __gateway.shutdown()
    # removing remote files
    if __remote_manager is not None:
        remote_deletable = __source_table.get_deletable("remote")
        for rd in remote_deletable:
            __remote_manager.delete_dataset(rd)
    # flushing the tmp files
    TempFileManager.flush_everything()


def is_backend_on():
    global __pythonManager
    return __pythonManager is not None


def get_python_api_package(gateway):
    return gateway.jvm.it.polimi.genomics.pythonapi


def start_gmql_manager(python_api_package):
    pythonManager = python_api_package.PythonManager
    pythonManager.startEngine()
    return pythonManager


def get_gateway():
    global __gateway

    if __gateway is None:
        # Starting the GMQL manager
        start()
        return __gateway
    else:
        return __gateway


def get_python_manager():
    global __pythonManager

    if __pythonManager is None:
        # Starting the GMQL manager
        start()
        return __pythonManager
    else:
        return __pythonManager


def __initialize_source_table():
    global __source_table
    from .dataset.loaders.Sources import SourcesTable
    __source_table = SourcesTable()


def get_source_table():
    global __source_table
    return __source_table


def __initialize_dependency_manager():
    global __dependency_manager
    __dependency_manager = DependencyManager()


def __initialize_session_manager():
    global __session_manager
    __session_manager = load_sessions()


def login():
    """ Enables the user to login to the remote GMQL service.
    If both username and password are None, the user will be connected as guest.
    """
    from .RemoteConnection.RemoteManager import RemoteManager
    global __remote_manager, __session_manager

    logger = logging.getLogger()
    remote_address = get_remote_address()
    res = __session_manager.get_session(remote_address)
    if res is None:
        # there is no session for this address, let's login as guest
        warnings.warn("There is no active session for address {}. Logging as Guest user".format(remote_address))
        rm = RemoteManager(address=remote_address)
        rm.login()
        session_type = "guest"
    else:
        # there is a previous session for this address, let's do an auto login
        # using that access token
        logger.info("Logging using stored authentication token")
        rm = RemoteManager(address=remote_address, auth_token=res[1])
        # if the access token is not valid anymore (therefore we are in guest mode)
        # the auto_login function will perform a guest login from scratch
        session_type = rm.auto_login(how=res[2])
    # store the new session
    __remote_manager = rm
    access_time = int(time.time())
    auth_token = rm.auth_token
    __session_manager.add_session(remote_address, auth_token, access_time, session_type)


def logout():
    """ The user can use this command to logout from the remote service

    :return: None
    """
    global __remote_manager
    __remote_manager.logout()


def execute_remote():
    global __remote_manager
    __remote_manager.execute_remote_all()


def get_remote_manager():
    """ Returns the current remote manager

    :return: a RemoteManager
    """
    global __remote_manager
    return __remote_manager


def get_session_manager():
    """ Returns the session manager of the current instance of the library

    :return: a SessionManager
    """
    global __session_manager
    return __session_manager


def __initialize_logger():
    log_fmt = '[PyGMQL] %(message)s'
    logging.basicConfig(level=logging.INFO, format=log_fmt)


def __check_dependencies():
    global __gmql_jar_path, __py4j_path
    if __gmql_jar_path is None:
        __gmql_jar_path = __dependency_manager.resolve_dependencies()
    if __py4j_path is None:
        __py4j_path = __check_py4j_backend()


def init_managers():
    __initialize_source_table()
    __initialize_session_manager()
    __initialize_dependency_manager()
    __initialize_logger()

    atexit.register(stop)
    signal.signal(signal.SIGINT, stop)

    __check_dependencies()



# from glob import glob
# from tqdm import tqdm
# import logging

# from ..parsers.MetadataParser import GenericMetaParser
# from . import generateNameKey
# import os
# import pandas as pd


# def load_meta_from_path(path):
#     meta_files = glob(pathname=path + '/*.gdm.meta')
#     parsed = []
#     parser = GenericMetaParser()
#     logger = logging.getLogger()
#     logger.debug("Loading meta data from path {}".format(path))
#     from ...settings import is_progress_enabled
#     for f in tqdm(meta_files, total=len(meta_files), disable=not is_progress_enabled()):
#         abs_path = os.path.abspath(f)
#         abs_path_no_meta = abs_path[:-5]
#         key = generateNameKey(abs_path_no_meta)
#         ps = parser.parse_metadata(abs_path)                # [(attr_name, value), ...]
#         ps = list(map(lambda x: (key, (x[0], x[1])), ps))   # [(id, (attr_name, value)),...]
#         # parsing
#         parsed.extend(ps)
#     return to_pandas(parsed)


# def to_pandas(meta_list):
#     # turn to dictionary
#     if len(meta_list) > 0:
#         meta_list = list(map(to_dictionary, meta_list))     # [{'id_sample': id, attr_name: value},...]
#         df = pd.DataFrame.from_dict(meta_list)
#         columns = df.columns
#
#         # grouping by 'id_sample'
#         g = df.groupby('id_sample')
#
#         logger = logging.getLogger()
#         logger.debug("dataframe construction")
#         result_df = pd.DataFrame()
#         from ...settings import is_progress_enabled
#         for col in tqdm(columns, total=len(columns), disable=not is_progress_enabled()):
#             if col != 'id_sample':
#                 result_df[col] = g[col].apply(to_list)
#     else:
#         result_df = pd.DataFrame()
#         result_df.index.name = "id_sample"
#     return result_df


# def to_dictionary(tuple):
#     return {"id_sample": tuple[0], tuple[1][0]: tuple[1][1]}
#
#
# def to_list(x):
#     l = list(x)
#     l = [a for a in l if not pd.isnull(a)]
#     return l


import unittest
import warnings


class TestPyGMQLImporting(unittest.TestCase):
    @staticmethod
    def test_import():
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=ImportWarning)
            import gmql as gl


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages
import os

with open(os.path.join(".", "gmql", "resources", "version"), "r") as f_ver:
    version = f_ver.read().strip()

with open(os.path.join(".", "gmql", "resources", "github_url"), "r") as f_ver:
    github_url = f_ver.read().strip()

setup(name='gmql',
      version=version,
      description='Python Library for data analysis based on GMQL',
      long_description="PyGMQL is a python module that enables the user "
                       "to perform operation on genomic data in a scalable way.",
      url=github_url,
      author='Luca Nanni, Pietro Pinoli, Arif Canakoglu, Stefano Ceri',
      author_email='luca.nanni@polimi.it',
      license='MIT',
      packages=find_packages(exclude=("gmql.ml",)),     # temporary
      scripts=['scripts/pygmql_login.py',
               'scripts/pygmql_win.bat',
               'scripts/pygmql.sh'],
      install_requires=['pandas',
                        'tqdm',
                        'numpy',
                        'py4j',
                        'requests',
                        'requests-toolbelt',
                        'strconv',
                        'typecheck-decorator',
                        'numpy', 'findspark',
                        'sphinx_rtd_theme'
                        # 'sklearn',
                        # 'pyclustering',
                        # 'matplotlib',
                        # 'scipy', 'wordcloud',
                        # fancyimpute'
                        ],
      classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
             'Topic :: Scientific/Engineering :: Information Analysis',
            'Programming Language :: Python :: 3'
      ],
      keywords="genomics big data pandas python",
      python_requires='>=3.0, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
      include_package_data=True,
      zip_safe=False)



from ..parsers import get_parsing_function
import pandas as pd
from tqdm import tqdm
from ..DataStructures import reg_fixed_fileds
from ...settings import get_regions_batch_size


# Loading strategy
loading_strategy = "single_core"
batch = get_regions_batch_size()


def load_regions(collected_result):
    # get the number of regions
    # n_regions = collected_result.getNumberOfRegions()
    # get the string delimiters
    regions_delimiter = collected_result.REGIONS_DELIMITER()
    values_delimiter = collected_result.VALUES_DELIMITER()
    end_of_stream = collected_result.END_OF_STREAM()
    # get how the strings are structured
    names, types = get_schema(collected_result)
    result = []

    from ...settings import is_progress_enabled
    if loading_strategy == 'single_core':
        bar = tqdm(disable=not is_progress_enabled(), desc='Collecting regions')
        region = collected_result.getRegionAsString(batch)
        while region != end_of_stream:
            region = region.split(regions_delimiter)
            bar.update(len(region))
            result.extend(map(lambda x: string_to_dictionary(x, values_delimiter, names, types), region))
            region = collected_result.getRegionAsString(batch)

        # # get the full string
        # regions_string = collected_result.getRegionAsString(none())
        # if regions_string:
        #     # convert to list of strings
        #     regions_string = regions_string.split(regions_delimiter)
        #     iterator = map(lambda x: string_to_dictionary(x, values_delimiter, names, types),
        #                    tqdm(regions_string, disable=not is_progress_enabled()))
        #     result.extend(iterator)
    # elif loading_strategy == 'multi_core':
    #     # number of divisions
    #     divisions = 10
    #     chunk_size = math.ceil(n_regions / divisions)
    #     p = pool.Pool(min(4, cpu_count()))
    #
    #     std_partial = partial(string_to_dictionary, values_delimiter=values_delimiter,
    #                           names=names, types=types)
    #
    #     for _ in tqdm(range(divisions), disable=not is_progress_enabled()):
    #         # get the full string
    #         regions_string = collected_result.getRegionsAsString(Some(chunk_size))
    #         if regions_string:
    #             # convert to list of strings
    #             regions_string = regions_string.split(regions_delimiter)
    #
    #             iterator = p.map(std_partial, regions_string)
    #             result.extend(iterator)
    #
    #     p.close()
    else:
        raise ValueError("Unknown loading mode ({})".format(loading_strategy))

    columns = reg_fixed_fileds + names
    if len(result) > 0:
        df = pd.DataFrame.from_dict(result)
        df = df.set_index(keys="id_sample", drop=True)
        df = df[columns]  # for ordering the columns
    else:
        df = pd.DataFrame(columns=columns)
        df.index.name = "id_sample"
    return df


def string_to_dictionary(region, values_delimiter, names, types):
    elements = region.split(values_delimiter)
    # print(elements)
    d = dict()
    d['id_sample'] = int(elements[0])
    d['chr'] = elements[1]
    d['start'] = int(elements[2])
    d['stop'] = int(elements[3])
    d['strand'] = elements[4]

    for i in range(0, len(names)):
        name = names[i]
        fun = types[i]
        element = elements[5 + i]
        if element != 'null':
            d[name] = fun(element)
        else:
            d[name] = None
    return d


def to_dictionary(region, names, types):
    d = dict()
    d['id_sample'] = int(region[0])
    d['chr'] = region[1]
    d['start'] = int(region[2])
    d['stop'] = int(region[3])
    d['strand'] = region[4]

    for i in range(0, len(names)):
        name = names[i]
        fun = types[i]
        d[name] = fun(region[5 + i])

    return d


def load_metadata(collected_result):
    meta_Java = collected_result.getMetadata()
    meta_list = []
    for meta in meta_Java:
        id = int(meta[0])
        name = meta[1]
        value = meta[2]
        meta_list.append({'id_sample' : id, name: value})

    if len(meta_list) > 0:
        df = pd.DataFrame.from_dict(meta_list)
        columns = df.columns

        # grouping by 'id_sample'
        g = df.groupby('id_sample')

        result_df = pd.DataFrame()
        from ...settings import is_progress_enabled
        for col in tqdm(columns, total=len(columns), disable=not is_progress_enabled()):
            if col != 'id_sample':
                result_df[col] = g[col].apply(to_list)
    else:
        result_df = pd.DataFrame()
        result_df.index.name = "id_sample"
    return result_df


def to_list(x):
    l = list(x)
    l = [a for a in l if not pd.isnull(a)]
    return l


def get_schema(collected_result):
    schema_Java = collected_result.getSchema()
    # schema_Java = [[name, type], [name, type], ...]
    names = []
    types = []
    for field in schema_Java:
        name = field[0]
        type_string = field[1]
        fun = get_parsing_function(type_string)

        names.append(name)
        types.append(fun)
    return names, types



from .dataset.GDataframe import from_pandas
from .dataset.loaders.Loader import load_from_path, load_from_remote, load_from_file
from .dataset import parsers, get_example_dataset
from .dataset.DataStructures.Aggregates import *
from .dataset.DataStructures.ExpressionNodes import *
from .dataset.DataStructures.GenometricPredicates import *
from .RemoteConnection.RemoteManager import RemoteManager
# from . import ml

from .settings import init_settings as __init_settings
from .managers import init_managers as __init_managers
from .settings import set_remote_address, set_meta_profiling, set_mode, set_progress, set_master, \
    set_local_java_options, set_spark_configs
from .managers import login, logout, get_remote_address, get_session_manager, set_backend_path, set_py4j_path, \
    get_configuration, get_remote_manager

__initiated = False


def __init():
    """ Initializes the library.
    :return: None
    """
    global __initiated
    if not __initiated:
        __init_settings()
        __init_managers()
        __initiated = True


__init()
